from __future__ import annotations
import asyncio

import logging
import json
import contextlib
import inspect

from discord.ext.commands import Bot, Cog
from typing import TYPE_CHECKING, Optional, Callable, ClassVar, TypeVar, Union, Type, Awaitable, Tuple, Any, Dict

from websockets.exceptions import ConnectionClosedError, ConnectionClosed
from websockets.server import WebSocketServerProtocol, WebSocketServer, serve

from .errors import InvalidReturn, NoEndpointFound, MulticastFailure, ServerAlreadyStarted
from .objects import ClientPayload


if TYPE_CHECKING:
    from typing_extensions import ParamSpec, TypeAlias
    
    P = ParamSpec("P")
    T = TypeVar("T")
    
    RouteFunc: TypeAlias = Callable[P, T]
    Handler = Callable[[WebSocketServerProtocol], Awaitable[Any]]

class Server:
    """
    The inter-process communication server that is connected with the bot.

    Parameters:
    ----------
    bot: `Bot`
        The bot instance.
    host: `str`
        The address for the server (the default host is 127.0.0.1 and you don't wanna change this in most cases).
    secret_key: `str`
        This string will be used as authentication password while taking requests.
    standard_port: `int`
        The port to run the standard server (the default is 1025).
    multicast_port: `int`
        The port to run the multicasting server (the default is 20000).
    do_multicast: `bool`
        Should the multicasting be allowed (this is enabled by default).
    logger: `Logger`
        You can specify the logger for logs related to the lib (the default is discord.ext.ipc.server)
    """

    endpoints: ClassVar[Dict[str, Tuple[RouteFunc, Type[ClientPayload]]]] = {}

    def __init__(
        self,
        bot: Bot,
        host: str = "127.0.0.1", 
        secret_key: Union[str, None] = None, 
        standard_port: int = 2505,
        multicast_port: int = 20000,
        do_multicast: bool = True,
        logger: Optional[logging.Logger] = None
    ) -> None:
        self.bot = bot
        self.host = host
        self.secret_key = secret_key
        self.standard_port = standard_port
        self.multicast_port = multicast_port
        self.do_multicast = do_multicast

        self.logger = logger or logging.getLogger(__name__)
        self.servers: Dict[str, WebSocketServer] = {}
        self.connection: Optional[Tuple[str, WebSocketServerProtocol]] = None

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} endpoints={len(self.endpoints)} started={self.started} standard_port={self.standard_port!r} multicast_port={self.multicast_port!r} do_multicast={self.do_multicast}>"

    def get_cls(self, func: RouteFunc) -> Union[Cog, Bot]:
        for cog in self.bot.cogs.values():
            if func.__name__ in dir(cog):
                return cog
        return self.bot

    @classmethod
    def route(cls, name: Optional[str] = None, multicast: Optional[bool] = True) -> Callable[[RouteFunc], RouteFunc]:
        """|method|

        Used to register a coroutine as an endpoint

        Parameters
        ----------
        name: :class:`str`
            The endpoint name. If not provided the method name will be used.
        multicast :class:`bool`
            Should the enpoint be avaiable for multicast or not. If this is set to False only standard connection can access it.
        """
        def decorator(func: RouteFunc) -> RouteFunc:
            payload = ClientPayload
            for annotation in func.__annotations__.values():
                if inspect.isclass(annotation) and issubclass(annotation, ClientPayload):
                    payload = annotation
                    break

            func.__multicast__ = multicast

            cls.endpoints[name or func.__name__] = (func, payload)
            return func
        return decorator

    @property
    def started(self) -> bool:
        return len(self.servers) > 0

    def is_secure(self, message: Union[str, bytes]) -> bool:
        data: Dict[str, Any] = json.loads(message)

        if (key := data.get("secret")):
            return str(key) == str(self.secret_key)
        return bool(self.secret_key is None)
    

    async def handle_request(self, websocket: WebSocketServerProtocol, message: Union[str, bytes], multucast: bool = True) -> None:
        payload: Dict[str, Union[str, int, None]] = {
            "decoding": None,
            "code": 200,
            "response": None
        }
        
        if not self.is_secure(message):
            payload["code"] = 403
            payload["error"] = "Unauthorized"
            payload["error_details"] = "You're trying to connect with an invalid secret key!"
            return await websocket.send(json.dumps(payload))

        data: Dict[str, Any] = json.loads(message)
        endpoint: str = data["endpoint"]

        if not (coro := self.endpoints.get(endpoint)):
            payload["code"] = 404
            payload["error"] = "Unknown endpoint!"
            payload["error_details"] = "The route that you're trying to call doesn't exist!"
            self.bot.dispatch("ipc_error", None, NoEndpointFound(endpoint, "The route that you're trying to call doesn't exist!"))
            return await websocket.send(json.dumps(payload))
        
        try:
            func = coro[0]

            if multucast and not func.__getattribute__("__multicast__"):
                payload["code"] = 500
                payload["error"] = "The requested route is not available for multicast connections!"
                payload["error_details"] = "This route can only be called with standart client!"
                self.bot.dispatch("ipc_error", endpoint, MulticastFailure(endpoint, "This route can only be called with standart client!"))
                return await websocket.send(json.dumps(payload))

            resp: Optional[Union[Dict, str]] = await func(self.get_cls(func), coro[1](data))
        except Exception as exc:
            payload["code"] = 500
            payload["error"] = "Unexpected error occurred while calling the route!"
            payload["error_details"] = str(exc)

            self.bot.dispatch("ipc_error", endpoint, exc)
            return await websocket.send(json.dumps(payload))
        
        if resp and not (isinstance(resp, Dict) or isinstance(resp, str)):
            payload["error"] = f"Expected type Dict or string as response, got {resp.__class__.__name__!r} instead!"
            payload["code"] = 500
            self.bot.dispatch("ipc_error", endpoint, InvalidReturn(endpoint, f"Expected type Dict or string as response, got {resp.__class__.__name__} instead!"))
            return await websocket.send(json.dumps(payload))
        
        if isinstance(resp, Dict):
            payload["decoding"] = "JSON"
            payload["response"] = json.dumps(resp)
        else:
            payload["response"] = resp
        
        return await websocket.send(json.dumps(payload))

    async def create_server(self, name: str, port: int, ws_handler: Handler) -> None:
        if name in self.servers:
            raise ServerAlreadyStarted(name, f"{name!r} is already started!")
        
        self.servers[name] = await serve(ws_handler, self.host, port)
        self.logger.debug(f"{name.title()} is ready for use!")

    async def standart_handler(self, websocket: WebSocketServerProtocol) -> None:
        id = websocket.request_headers["ID"]
        
        try:
            if self.connection:
                if not self.connection[0] == id:
                    resp = json.dumps({
                        "error": "Connection already reserved!",
                        "details": self.connection[0],
                        "code": 422
                    })
                    
                    return await websocket.send(resp)
            else:
                self.logger.debug(f"New connection created: {id}")
                self.connection = id, websocket
        
            async for message in websocket:
                with contextlib.suppress(ConnectionClosedError, ConnectionClosed):
                    await asyncio.create_task(self.handle_request(websocket, message, False))
            
        except (ConnectionClosedError, ConnectionClosed):
            self.logger.debug(f"Connection closed by the client: {self.connection[0]}") # type: ignore
            self.connection = None
    
    async def multicast_handler(self, websocket: WebSocketServerProtocol) -> None:
        with contextlib.suppress(ConnectionClosedError, ConnectionClosed):
            async for message in websocket:
                await asyncio.create_task(self.handle_request(websocket, message, True))

    async def start(self) -> None:
        """|coro|
        
        Starts all necessary processes for the servers and runners to work properly

        """
        await self.create_server("standard", self.standard_port, self.standart_handler)

        if self.do_multicast:
            await self.create_server("mutlicast", self.multicast_port, self.multicast_handler)

        self.bot.dispatch("ipc_ready")
    
    async def stop(self) -> None:
        """|coro|

        Takes care of shutting down the server(s)

        """
        for name, server in self.servers.items():
            server.close()
            await server.wait_closed()
            self.logger.info(f"{name!r} server has been stopped!")
        
        self.servers = {}


from __future__ import annotations

import asyncio
import logging
import json
import uuid
import warnings

from typing import Optional, Union, Any, Dict
from websockets.client import connect, WebSocketClientProtocol 
from .objects import ServerResponse
from .errors import ServerTimeout

class Client:
    """
    The client that connects to a Server and sends requests to it

    Parameters:
    ----------
    host: `str`
        The address for the server (the default host is 127.0.0.1 and you don't wanna change this in most cases).
    secret_key: `str`
        This string will be used as authentication password while making requests.
    standard_port: `int`
        The port to run the standard server (the default is 2505).
    multicast_port: `int`
        The port to run the multicasting server (the default is 20000).
    do_multicast: `bool`
        Should the client perform standard or multicast connection (this is enable by default)
    timeout: `Optional[int]`
        How long should each request take before raising `Timeout` exceptiom
    """

    def __init__(
        self,
        host: str = "127.0.0.1",
        secret_key: Union[str, None] = None,
        standard_port: int = 2505,
        multicast_port: int = 20000,
        do_multicast: bool = True,
        timeout: Optional[int] = None
    ) -> None:
        self.host = host
        self.standard_port = standard_port
        self.secret_key = secret_key
        self.multicast_port = multicast_port
        self.do_multicast = do_multicast
        self.timeout = timeout

        self.id = uuid.uuid4()
        self.logger = logging.getLogger(__name__)
        self.connection: Optional[WebSocketClientProtocol] = None
        self.lock = asyncio.Lock()

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} standard_port={self.standard_port!r} multicast_port={self.multicast_port!r} do_multicast={self.do_multicast}>"

    @property
    def url(self) -> str:
        if self.do_multicast:
            return f"ws://{self.host}:{self.multicast_port}"
        return f"ws://{self.host}:{self.standard_port}"

    async def request(self, endpoint: str, **kwargs: Any) -> Optional[ServerResponse]:
        """|coro|
        
        Makes a request to the server URL.

        ----------
        endpoint: `str`
            The endpoint to request on the server
        **kwargs: `Any`
            The data for the endpoint
        """
        
        if not self.do_multicast:
            warnings.warn("Multicast will replace normal connection in the future!", PendingDeprecationWarning)

            if not self.connection:
                self.connection = await connect(self.url, extra_headers={"ID": str(self.id)})
        
            await self.connection.send(json.dumps({
                "endpoint": endpoint,
                "secret": str(self.secret_key),
                "kwargs": {
                    **kwargs
                },
            }))
            
            
            async with self.lock:
                try:
                    return ServerResponse(await asyncio.wait_for(self.connection.recv(), self.timeout)) # type: ignore
                except asyncio.TimeoutError:
                    raise ServerTimeout(f"The server response took longer than the maximum timeout!", self.timeout) # type: ignore
        
        async with connect(self.url) as conn:
            await conn.send(json.dumps({
                "endpoint": endpoint,
                "secret": str(self.secret_key),
                "kwargs": {
                    **kwargs
                },
            }))
            
            try:
                return ServerResponse(await asyncio.wait_for(conn.recv(), self.timeout)) # type: ignore
            except asyncio.TimeoutError:
                raise ServerTimeout(f"The server response took longer than the maximum timeout!", self.timeout) # type: ignore



import json
import enum
from typing import Dict, Any, Optional, Union

class StatusEnum(enum.Enum):
    OK = 200
    FORBIDDEN = 403
    NOT_FOUND= 404
    INTERNAL_ERROR = 500
    
    def __str__(self) -> str:
        return str(self.value)

class ClientPayload:
    """
    The base class for the payload which is sent to the endpoint
    when the call is made. 
    
    This can be subclassed and custom payload
    can be used. If you do not Typehint the function with the custom
    payload then it will automatically use this base payload,
    but keys and values can be accessed like a dictionary or using `X.y`.

    Parameters:
    ----------
    payload: Dict`
        The payload to be converted.

    Attributes
    ----------
    length: `int`
        The length of the payload.
    endpoint: `str`
        The endpoint which was called.
    data: `Dict`
        The kwargs from the payload.
    """

    def __init__(self, payload: Dict[str, Any]):
        self.payload = payload
        self.length: int = len(payload)
        self.endpoint: Optional[str] = payload.get("endpoint")
        self.data: Optional[Dict[str, Any]] = payload.get("kwargs")

    def __getitem__(self, __k: str):
        return self.data[__k] # type: ignore

    def __contains__(self, __o: object) -> bool:
        return __o in self.data or __o in self.data.values() # type: ignore

    def __getattribute__(self, __name: str) -> Any:
        try:
            return object.__getattribute__(self, __name)
        except AttributeError:
            try:
                return self.data[__name] # type: ignore
            except KeyError:
                raise AttributeError(__name)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} length={self.length} endpoint={self.endpoint!r}>"

    @property
    def raw(self) -> Dict:
        return self.payload

    def items(self):
        """|method|

        Returns the payload in the form of dictionary items.
        """
        return self.payload.items()

class ServerResponse:
    """
    The class when getting response for the Server

    Parameters:
    ----------
    payload: `str`
        The payload to be converted.

    Attributes
    ----------
    response: `Dict | str`
        Decoded response that is ready for use..
    error: `Dict | None`
        Returns dict with exception information (if any).
    status: `StatusEnum`
        Raw status code converted to readable data.
    """
    def __init__(self, payload: str):
        self.data: Dict[str, Any] = json.loads(payload)
        self.decoding: str = self.data["decoding"]

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} response={self.response} status={self.status.name}>"
    
    @property
    def response(self) -> Union[Dict, str]:
        """|property|
        
        Decoded response that is ready for use.

        """
        if self.decoding == "JSON":
            return json.loads(self.data["response"])
        return self.data["response"]

    @property
    def error(self) -> Optional[Dict[str, Any]]:
        """|property|

        Optionally returns any errors that from the server side.

        """
        if (error := self.data.get("error")):
            return {
                "error": error,
                "status": self.status,
                "details": self.data.get("error_details"),
            }

    @property
    def status(self) -> StatusEnum:
        """|property|
        
        The status code after being converted to `enum.Enum`

        """
        return StatusEnum(self.data.get("code"))



import discord

if discord.version_info.major < 2:
    raise RuntimeError("You must have py-cord (v2.0 or greater) to use this library.")

__title__ = "pycord-ipc"
__author__ = "Sampath"
__license__ = "GNU GENERAL PUBLIC LICENSE"
__copyright__ = "Copyright (C) 2023 Sampath"
__version__ = "1.0.0"


from .errors import BaseException, NoEndpointFound, MulticastFailure, InvalidReturn, ServerAlreadyStarted
from .client import Client
from .server import Server
from .objects import ClientPayload, ServerResponse




class BaseException(Exception):
    """Common base class for all exceptions"""
    def __init__(self, name: str, details: str) -> None:
        self.name = name
        self.details = details

class NoEndpointFound(BaseException):
    """Raised when trying to request an unknown endpoint"""

class MulticastFailure(BaseException):
    """Raised when calling route that is not multicasted"""

class InvalidReturn(BaseException):
    """Raiseed when getting un-serializable objects as response from a route"""
    pass

class ServerAlreadyStarted(BaseException):
    """Raise trying to start already running server"""

class ServerTimeout(BaseException):
    def __init__(self, details: str, timeout: int) -> None:
        self.details = details
        self.timeout = timeout